module Lib
    ( someFunc
    ) where

import qualified Data.Map as Map
import Control.Parallel (par, pseq)

-- Given a number n, makes a nxn matrix (2D) with all elements set to 0, using a map where the key is (row, col) and the value is 0
generateMatrix :: Int -> Map.Map (Int, Int) Int
generateMatrix n = Map.fromList [((i, j), 0) | i <- [0..n-1], j <- [0..n-1]]

getCoordsFromIndex :: Int -> Int -> (Int, Int)
getCoordsFromIndex index n = (index `div` n, index `mod` n)

-- Given a matrix board, place a queen at the given index, return the new matrix
placeQueen :: Int -> Map.Map (Int, Int) Int -> Int -> Map.Map (Int, Int) Int
placeQueen index board n = Map.insert (getCoordsFromIndex index n) 1 board

-- Validate all rows: Every row must sum to 1
validateRows :: Map.Map (Int, Int) Int -> Int -> Bool
validateRows board n = all (\r -> sum [board Map.! (r, c) | c <- [0..n-1]] == 1) [0..n-1]

-- Validate all columns: Every column must sum to 1
validateCols :: Map.Map (Int, Int) Int -> Int -> Bool
validateCols board n = all (\c -> sum [board Map.! (r, c) | r <- [0..n-1]] == 1) [0..n-1]

-- Validate all diagonals: Every diagonal must sum to 1 or less
-- validateDiagonals :: Map.Map (Int, Int) Int -> Int -> Bool
-- validateDiagonals board n = all (\d -> sum [board Map.! (r, c) | (r, c) <- [(r, d - r) | r <- [0..n-1], d - r >= 0], r < n, d - r < n] == 1) [0..2*n-2]

validateDiagonals :: Map.Map (Int, Int) Int -> Int -> Bool
validateDiagonals board n =
    let mainDiagonal = sum [Map.findWithDefault 0 (i, i) board | i <- [1..n]]
        antiDiagonal = sum [Map.findWithDefault 0 (i, n - i + 1) board | i <- [1..n]]
    in mainDiagonal == 1 && antiDiagonal == 1

-- Validate that there are n queens on the board in total
validateNumQueens :: Map.Map (Int, Int) Int -> Int -> Bool
validateNumQueens board n = sum (Map.elems board) == n

-- Given a matrix board represented as a map (Int, Int) Int, validate it (return 1 if valid, 0 otherwise)
-- The sum of every row must be 1
-- The sum of every column must be 1
-- The sum of every diagonal must be 1 (and every anti-diagonal must be 1)
validateBoard :: Map.Map (Int, Int) Int -> Int -> Int
validateBoard board n
    | not (validateRows board n) = 0
    | not (validateCols board n) = 0
    | not (validateDiagonals board n) = 0
    | not (validateNumQueens board n) = 0
    | otherwise = 1


-- V0: BASE NO PARALLEL
-- Given a current index, a number of the remaining queens, and the matrix board as a hashmap,
-- return the number of ways you can place the remaining queens
-- solveNQueens :: Int -> Int -> Map.Map (Int, Int) Int -> Int
-- solveNQueens n index board
--     | index == (Map.size board) = validateBoard board n
--     | otherwise = solveNQueens n (index + 1) board + solveNQueens n (index + 1) (placeQueen index board n)


-- V1
-- Naive Parallel version of solveNQueens
-- solveNQueens :: Int -> Int -> Map.Map (Int, Int) Int -> Int
-- solveNQueens n index board
--     | index == (Map.size board) = validateBoard board n
--     | otherwise = solution1 `par` solution2 `par` (solution1 + solution2) -- not sure if this pseq is really necessary, probs some overhead
--     where
--         solution1 = solveNQueens n (index + 1) board
--         solution2 = solveNQueens n (index + 1) (placeQueen index board n)


-- V2: Naive Parallel version of solveNQueens with pseq to reduce spark overflow
-- solveNQueens :: Int -> Int -> Map.Map (Int, Int) Int -> Int
-- solveNQueens n index board
--     | index == (Map.size board) = validateBoard board n
--     | otherwise = solution1 `par` (solution2 `pseq` (solution1 + solution2)) -- pseq uses see notes
--     where
--         solution1 = solveNQueens n (index + 1) board
--         solution2 = solveNQueens n (index + 1) (placeQueen index board n)


-- V3 Sequential portion (to be used by V3.1, V3.2, V3.3). Also new input of depth
solveNQueensSequential :: Int -> Int -> Map.Map (Int, Int) Int -> Int
solveNQueensSequential n index board
    | index == (Map.size board) = validateBoard board n
    | otherwise = solveNQueensSequential n (index + 1) board + solveNQueensSequential n (index + 1) (placeQueen index board n)

-- Updated someFunc to take depth as input
someFunc :: Int -> Int -> String
someFunc n depth = "Answer: " ++ show (solveNQueens depth n (generateMatrix n) 0)


-- V3.1: Parallel up until certain depth.
-- Parallel version: handles parallelism up to a fixed depth
solveNQueens :: Int -> Int -> Map.Map (Int, Int) Int -> Int -> Int
solveNQueens 0 n board index = solveNQueensSequential n index board
solveNQueens _ n board index | index == (n * n) = validateBoard board n
solveNQueens depth n board index =
    solution1 `par` solution2 `pseq` (solution1 + solution2)
  where
    solution1 = solveNQueens (depth - 1) n board (index + 1)
    solution2 = solveNQueens (depth - 1) n (placeQueen index board n) (index + 1)



-- someFunc must take in an Int and return a String (that really just says how many
-- different ways you can solve the n-queens problem for a given n, the input)
-- someFunc :: Int -> String
-- someFunc n = "Answer: " ++ show (solveNQueens n 0 (generateMatrix n))
